(require kodhy.macros)

(import
  [math [*]]
  random
  os.path
  [numpy :as np]
  [pandas :as pd]
  [sklearn.linear-model :as sklm]
  [sklearn.ensemble :as sken]
  [scipy.stats :as scist]
  fancyimpute
  [kodhy.util [
    all-unique? ret cached-eval pairs
    bin-labels choose-labeled-cv-folds with-1o-interacts]])

(setv cache-dir (os.path.join (os.path.expanduser "~")
  ".daylight" "py-cache" "Kodi-projects" "pelt"))

(defn simple-impute [x]
  ; Replace missing values in x.
  (setv x (.copy x))
  (for [col-i (range (second x.shape))]
    (setv col (geta x : col-i))
    (setv (get x (, (np.isnan col) col-i))
      (if (> (len (np.unique (get col (~ (np.isnan col))))) 2)
        ; If the column has more than 2 unique values, use
        ; the median.
        (np.nanmedian col)
        ; Otherwise, this column may be for a binary variable, so
        ; use the mode. (Which is equal to the median except for the
        ; edge case of an even split between 2 distinct values.)
        (float (first (scist.mode col))))))
  x)

(defn soft-impute [x]
  (when (np.any (np.isnan x))
    (setv binary-col (amap
      (<= (len (np.unique (geta x (~ (np.isnan (geta x : it))) it))) 2)
      (range (second x.shape))))
    (setv x (.complete (kwc fancyimpute.SoftImpute :+verbose) x))
    ; In binary columns, round values to 0 or 1.
    (for [[binary? col-i] (zip binary-col (range (second x.shape)))]
      (unless binary?
        (continue))
      (setv (geta x : col-i) (np.maximum 0 (np.minimum 1
      (np.round (geta x : col-i)))))))
  x)

(defn standardize [x]
  (setv x (.copy x))
  (for [col-i (range (second x.shape))]
    (setv col (geta x : col-i))
    (when (> (len (np.unique (get col (~ (np.isnan col))))) 2)
      ; Only standardize features that aren't binary.
      (setv (geta x : col-i) (/ (- col (np.nanmean col)) (* 2 (np.nanstd col))))))
  x)

(defn model-trivial [x-train y-train x-test]
  (np.full-like x-test (np.median y-train)))

(defn model-ols [x-train y-train x-test] (block
  (unless (> (first x-train.shape) (+ (second x-train.shape) 1))
    ; The system isn't overdetermined. Fall back to the trivial model.
    (ret (model-trivial x-train y-train x-test)))
  (setv m (sklm.LinearRegression))
  (.fit m x-train y-train)
  (.predict m x-test)))

(defn model-enet [x-train y-train x-test]
  (setv m (kwc sklm.ElasticNetCV
    :l1-ratio [.01 .1 .25 .5 .75 .9 .95 .99 1]
    :cv 5
    :n-jobs 3))
  (.fit m x-train y-train)
  (.predict m x-test))

(defn model-enet-1oi [x-train y-train x-test]
  (setv [x-train x-test] (np.vsplit
    (with-1o-interacts (np.vstack [x-train x-test]))
    [ (first x-train.shape) ]))
  (model-enet x-train y-train x-test))

(defn model-rf [x-train y-train x-test]
  (setv m (kwc sken.RandomForestRegressor
    :n-estimators 500))
  (.fit m x-train y-train)
  (.predict m x-test))

(defn assess-predictions [y y-pred]
  (np.mean (np.abs (- y y-pred))))
    ; Mean absolute error.

(defn prep-data [x y y-range]

  ; Convert the data, and impute missing values in x.
  (setv s (pd.Series x.index))
  (setv x (soft-impute (standardize (.as-matrix x))))
  (setv y (.as-matrix y))

  ; Exclude rows for subjects missing on y.
  ; (We impute on the IV side, but simply omit on the DV side.)
  (setv nonnull-ix (~ (np.isnan y)))
  (setv s (get s nonnull-ix))
  (setv x (get x nonnull-ix))
  (setv y (get y nonnull-ix))

  (assert (& (>= (.min y) (first y-range)) (<= (.max y) (second y-range))))

  ; Choose cross-validation folds.
  (setv folds (block
    (setv labels (amap (get treatment-groups it) s))
    (setv n-folds 10)
    (setv n-groups (len (set labels)))
    (cond
      [(< (len s) (* n-folds 2))
        ; Use a leave-one-out method. (Each subject gets their
        ; own fold.)
        (ret (amap [it] s))]
      [(< n-groups 6)
        (raise ValueError)]
      [(< n-groups 12)
        (setv n-folds 4)])
    (setv max-bin-size (int (ceil
      (+ (/ (len labels) n-folds) (/ n-folds 2)))))
    (setv poss (kwc cached-eval
      :cache-dir cache-dir
      :key (str (.format "tgroup bin-labels {} {} {}"
        n-folds max-bin-size (.join " " (sorted s))))
      :f (fn [] (bin-labels labels n-folds max-bin-size))))
    (for [p poss]
      (print "Possible label bin sizes:" (sorted (map sum p))))
    (setv rstate (random.getstate))
    (try
      (do
        (random.seed 23)
        (ret (choose-labeled-cv-folds s labels poss)))
      (finally
        (random.setstate rstate)))))

  {"s" s "x" x "y" y "folds" folds})

(defn primary-predict [model y-range s x y folds]
  ; Expects input from prep-data.
  (setv [y-min y-max] y-range)
  (setv y-pred (np.empty-like y))
  (for [[i fold] (enumerate folds)]
    (print "Fold" i)
    (setv ix (. (.isin s fold) values))
    (setv x-train (get x (~ ix)))
    (setv y-train (get y (~ ix)))
    (setv x-test (get x ix))
    (setv (get y-pred ix) (np.minimum y-max (np.maximum y-min
      (model x-train y-train x-test)))))
  (assess-predictions y y-pred))

(defn primary-predict-allmodels [x reduced-ivs treatment-iv y y-range]
  (setv x (.sort-index x))
  (setv y (.sort-index y))
  (assert (all-unique? x.columns))
  (assert (all (lc [[xiv yiv] (zip x.index y.index)] (= xiv yiv))))
  (assert (.issubset (set reduced-ivs) x.columns))
  (assert (all (amap (.any (.notnull (getl x : it))) x.columns)))
  (setv v (kwc cached-eval
    :cache-dir cache-dir
    :key (str (.format "primary-predict-allmodels of {} given [{}] [{}] of [{}]"
      y.name (.join " " x.columns) (.join " " (sorted reduced-ivs)) (.join " " x.index)))
    :f (fn [] 
      (setv dat (prep-data x y y-range))
      (setv dat-reduced (prep-data (getl x : (+ reduced-ivs (if treatment-iv [treatment-iv] []))) y y-range))
      (setv dat-treatment (.copy dat))
      (when treatment-iv
        ; In dat-treatment, add 'x' columns for interactions with
        ; the treatment.
        (setv x2 (np.delete (.copy (get dat-treatment "x"))
          (.index (list x.columns) treatment-iv)
          1))
        (setv (geta x2 (.as-matrix (= (getl x (get dat-treatment "s") treatment-iv) 0)) :) 0)
        (setv (get dat-treatment "x") (np.hstack [(get dat-treatment "x") x2]))
        (assert (= (second (. (get dat-treatment "x") shape)) (dec (* (len x.columns) 2)))))
      (defn f [model data]
        (fn [] (apply primary-predict [model y-range] data)))
      (setv p (pairs
        :trivial (f model-trivial dat)
        :ols-reduced (f model-ols dat-reduced)
        ; :ols (f model-ols dat)
        :enet-reduced (f model-enet dat-reduced)
        :enet (f model-enet dat-treatment)
        :enet-1oi (f model-enet-1oi dat)
        :rf (f model-rf dat)))
      (setv results (kwc pd.Series :index (amap (first it) p) (amap
        (do
          (print "Fitting" (first it) "for" y.name)
          ((second it)))
        p)))
      (pd.concat [
        (kwc pd.Series :index ["n"] (len (get dat "s")))
        results]))))
  (setv best-baseline (.min (getl v ["trivial" "ols-reduced" "enet-reduced"])))
  (setv best-critical (.min (getl v ["enet" "rf" "enet-1oi"])))
  (setv (getl v "Improve-diff") (- best-baseline best-critical))
  (setv (getl v "Improve-ratio") (/ best-critical best-baseline))
  v)

(defn primary-predict-multi-dv [subject-ids x grouping-ivs treatment-iv other-ivs ys]
  (setv [y-names y-vectors y-ranges reduced-ivs-each] (apply zip ys))
  (setv d (kwc pd.concat :axis 1 (lc [[y yr riv] (zip y-vectors y-ranges reduced-ivs-each)]
    (primary-predict-allmodels
      (getl x subject-ids (+
        grouping-ivs other-ivs (if treatment-iv [treatment-iv] [])))
      (+ grouping-ivs riv)
      treatment-iv
      (getl y subject-ids)
      yr))))
  (setv d.columns y-names)
  d)
